/*
 * This file belongs to the XQuare distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquare.org.
 */

package org.xquark.mediator.plan;

import java.util.ArrayList;

import org.xquark.datastructure.DOMUtils;
import org.xquark.datastructure.Tuple;
import org.xquark.mediator.DOMUtils.DOMEval;
import org.xquark.mediator.algebra.Algebra;
import org.xquark.mediator.decomposer.Utils;
import org.xquark.mediator.runtime.MediatorException;
import org.xquark.xpath.datamodel.TypedNode;
import org.xquark.xquery.parser.Value;
import org.xquark.xquery.parser.Variable;
import org.xquark.xquery.parser.XQueryException;
import org.xquark.xquery.parser.XQueryExpression;

public class OpValue extends ZeroOp {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.12 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";

	private XQueryExpression condition = null;
	private boolean onlyVar = false;
	// ***********************************************************************
	// * INITIALIZATION
	// ***********************************************************************
	/**
	 *
	 */
	public OpValue(ExecutionPlan plan, Algebra algebra, boolean hasIdentifier) throws XQueryException {
		super(plan, algebra.getExpression());
		/*
		this.resetPaths();
		this.addPath(var);
		size = 1;
		*/
		Variable var = (Variable) algebra.getVariables().get(0);
		// added for external variable special case
		if (expression == null) {
		    expression = var;
		}
		if (algebra.getPaths().isEmpty()) {
			addPath(var);
			size = 1;
			onlyVar = true;
		} else {
			ArrayList tmpPaths = algebra.getPaths();
			addPaths(tmpPaths);
			size = tmpPaths.size();
			if (tmpPaths.size() == 1 && ((XQueryExpression)tmpPaths.get(0)).getStringValue().equals(var.getStringValue()))
				onlyVar = true;
		}
		idsize = 0;
		if (plan.getIdVars().contains(var) || hasIdentifier)
			idsize++;
		this.isLet(algebra.isLet());
	}

	//	public OpValue(ExecutionPlan plan, Algebra algebra, XQueryExpression expression, Variable var, XQueryExpression condition) throws XQueryException {
	//		super(plan, expression);
	//		this.condition = condition;
	//		/*
	//		this.resetPaths();
	//		this.addPath(var);
	//		size = 1;
	//		*/
	//		idsize = 0;
	//	}

	public XQueryExpression getcondition() {
		return this.condition;
	}

	public Operator addCondition(ExecutionPlan plan, XQueryExpression expr) throws XQueryException {
		if (condition == null) {
			this.condition = expr;
			return this;
		} else
			return new OpRestrict(plan, expr, this);
	}

	public boolean isOnlyVar() {
		return onlyVar;
	}

	// #############################################################################
	// VISITOR STUFF
	// #############################################################################

	public void accept(OperatorVisitor visitor) throws XQueryException {
		visitor.visit(this);
	}

	// ***********************************************************************
	// * EXECUTE QUERY
	// ***********************************************************************
	/**
	 *
	 */
	protected ResultSet getResultSet(DynamicContext context) throws XQueryException {
		return new ValueResultSet(this, context);
	}

	// **********************************************************************
	// * OPTIMIZE
	// **********************************************************************
	/**
	 * Return all the sources depending of this node.
	 */
	public ArrayList getSources() {
		if (sources == null) {
			sources = new ArrayList();
		}
		return sources;
	}

	// **********************************************************************
	// * DEBUG
	// **********************************************************************
	/**
	 *
	 */
	public String toCompleteString(int indent) {
		StringBuffer buf = new StringBuffer();
		buf.append(Utils.makeIndent(indent) + "<" + getClass().getName() + " isLet=\"" + isLet() + "\">\n");
		buf.append(Utils.makeIndent(indent + 1) + "<Expression>\n");
		buf.append(Utils.makeIndent(indent + 2) + expression + "\n");
		buf.append(Utils.makeIndent(indent + 1) + "</Expression>\n");
		buf.append(Utils.makeIndent(indent + 1) + "<condition>\n");
		buf.append(Utils.makeIndent(indent + 2) + condition + "\n");
		buf.append(Utils.makeIndent(indent + 1) + "</condition>\n");
		//		if (variables != null) {
		//			buf.append(Utils.makeIndent(indent + 1) + "<Variables>\n") ;
		//			for (int i = 0 ; i < variables.size() ; i ++) {
		//				buf.append(Utils.makeIndent(indent + 2) + "<Variable>" + variables.get(i) + "</Variable>\n") ;
		//			}
		//			buf.append(Utils.makeIndent(indent + 1) + "</Variables>\n") ;
		//		}
		if (paths != null) {
			buf.append(Utils.makeIndent(indent + 1) + "<Paths>\n");
			for (int i = 0; i < paths.size(); i++) {
				buf.append(Utils.makeIndent(indent + 2) + "<Path>" + paths.get(i) + "</Path>\n");
			}
			buf.append(Utils.makeIndent(indent + 1) + "</Paths>\n");
		}
		buf.append(Utils.makeIndent(indent) + "</" + getClass().getName() + ">\n");
		return buf.toString();
	}

}

class ValueResultSet extends ZeroResultSet {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.12 $";
	private static final String RCSName = "$Name: XQ1_1_1 $";
	// ***********************************************************************
	// * INITIALIZATION
	// ***********************************************************************
	/**
	 *
	 */
	public ValueResultSet(OpValue operator, DynamicContext context) throws XQueryException {
		super(operator, context);
	}

	// ***********************************************************************
	// * EVALUATE IMPLEMENTATION
	// ***********************************************************************
	/**
	 * In a cartesian product, we can all do in parallel, so we
	 * must generate all results of both source and then make the cartesian
	 * product.
	 *
	 * @param non_blocking if the evaluation is blocking, we must
	 *        generate all the results before passing.
	 */

	//private static EvaluationVisitor visitor = null;

	protected void evaluate(boolean non_blocking) throws XQueryException {
		//EvaluationVisitor visitor = null;
		XQueryExpression condition = ((OpValue) operator).getcondition();
		Tuple tuple;
		if (expression instanceof Value) {
			if (condition != null) {
				if (!DOMEval.evalCondition(context, condition, null)) {
					setFinishedWhenEmpty();
					return;
				}
			}
			tuple = new Tuple(1, this.operator.getIdSize());
			tuple.addNodeAtIndex(0, DOMUtils.makeTextNode((Value) expression, this.operator.getPlan().getSchemaManager()));
			if (this.operator.getIdSize() > 0)
				tuple.fillIdentifiers();
			buftuples.add(tuple);
		} else {
			ArrayList resnodes = DOMEval.evalNodes(context, expression, null);
			if (resnodes != null && !resnodes.isEmpty()) {
				if (operator.isLet()) {
					tuple = new Tuple(operator.getSize(), operator.getIdSize());
					if (((OpValue) operator).isOnlyVar()) {
						tuple.addNodesAtIndex(0, resnodes);
						tuple.setParent(buftuples);
						if (this.operator.getIdSize() > 0)
							tuple.fillIdentifiers();
						if (condition == null || DOMEval.evalCondition(context, condition, tuple))
							buftuples.add(tuple);
					} else {
						boolean doAdd = false;
						for (int i = 0; i < resnodes.size(); i++) {
							ArrayList nodelist = new ArrayList(1);
							nodelist.add(resnodes.get(i));
							// get paths but test if path is expression !!!
							for (int j = 0; j < paths.size(); j++) {
								if (paths.get(j).toString().equals(expression.getStringValue())) {
									tuple.addNodeAtIndex(j, (TypedNode) resnodes.get(i));
									doAdd = true;
								}
								else {
									ArrayList reslist = DOMUtils.getSteps(nodelist, (XQueryExpression) paths.get(j));
									if (reslist != null) {
										tuple.addNodesAtIndex(j, reslist);
										doAdd = true;
									}
								}
							}
						}
						if (doAdd && (condition == null || DOMEval.evalCondition(context, condition, tuple))) {
							tuple.fillIdentifiers();
							buftuples.add(tuple);
						}
					}
			
				} else {
					if (((OpValue) operator).isOnlyVar()) {
						for (int i = 0; i < resnodes.size(); i++) {
							tuple = new Tuple(1, this.operator.getIdSize());
							tuple.addNodeAtIndex(0, (TypedNode) resnodes.get(i));
							tuple.setParent(buftuples);
							if (this.operator.getIdSize() > 0)
								tuple.fillIdentifiers();
							if (condition == null || DOMEval.evalCondition(context, condition, tuple))
								buftuples.add(tuple);
						}
					} else {
						for (int i = 0; i < resnodes.size(); i++) {
							ArrayList nodelist = new ArrayList(1);
							nodelist.add(resnodes.get(i));
							tuple = buftuples.newTuple();
							boolean doAdd = false;
							// get paths but test if path is expression !!!
							for (int j = 0; j < paths.size(); j++) {
								if (paths.get(j).toString().equals(expression.getStringValue())) {
									tuple.addNodeAtIndex(j, (TypedNode) resnodes.get(i));
									doAdd = true;
								}
								else {
									ArrayList reslist = DOMUtils.getSteps(nodelist, (XQueryExpression) paths.get(j));
									if (reslist != null) {
										tuple.addNodesAtIndex(j, reslist);
										doAdd = true;
									}
								}
							}
							if (doAdd && (condition == null || DOMEval.evalCondition(context, condition, tuple))) {
								tuple.fillIdentifiers();
								buftuples.add(tuple);
							}
						}
					}
				}
			} else
				throw new MediatorException("OpValue : Could not evaluate expression : " + expression);
		}

		setFinishedWhenEmpty();
	}
}
